import os

import pytest
from dotenv import load_dotenv
from langchain_core.documents import Document
from langchain_experimental.graph_transformers import LLMGraphTransformer
from langchain_openai import ChatOpenAI

from langchain_memgraph.graphs.memgraph import MemgraphLangChain

# Load environment variables from .env
load_dotenv()


@pytest.fixture
def memgraph_connection():
    """Setup Memgraph connection fixture."""
    url = os.getenv("MEMGRAPH_URL", "bolt://localhost:7687")
    username = os.getenv("MEMGRAPH_USER", "")
    password = os.getenv("MEMGRAPH_PASSWORD", "")

    graph = MemgraphLangChain(
        url=url, username=username, password=password, refresh_schema=False
    )
    yield graph

    # Cleanup: clear the database after test
    graph.query("MATCH (n) DETACH DELETE n")


@pytest.fixture
def llm_transformer():
    """Setup LLM transformer fixture."""
    os.environ["OPENAI_API_KEY"] = os.getenv("OPENAI_API_KEY", "")
    llm = ChatOpenAI(temperature=0, model_name="gpt-4-turbo")
    return LLMGraphTransformer(llm=llm)


def test_ingest_and_verify(memgraph_connection, llm_transformer):
    """Test to ingest a document and check nodes/edges in Memgraph."""
    text = """
        Charles Robert Darwin was an English naturalist, geologist, and biologist,
        widely known for his contributions to evolutionary biology. His proposition that
        all species of life have descended from a common ancestor is now generally
        accepted and considered a fundamental scientific concept. In a joint
        publication with Alfred Russel Wallace, he introduced his scientific theory that
        this branching pattern of evolution resulted from a process he called natural
        selection, in which the struggle for existence has a similar effect to the
        artificial selection involved in selective breeding. Darwin has been
        described as one of the most influential figures in human history and was
        honoured by burial in Westminster Abbey.
    """

    # Convert document to graph structure
    documents = [Document(page_content=text)]
    graph_documents = llm_transformer.convert_to_graph_documents(documents)

    # Ingest graph data into Memgraph
    memgraph_connection.add_graph_documents(graph_documents)

    # Verify nodes and edges were created
    node_count = memgraph_connection.query("MATCH (n) RETURN count(n) AS count")[0][
        "count"
    ]
    edge_count = memgraph_connection.query("MATCH ()-[r]->() RETURN count(r) AS count")[
        0
    ]["count"]

    assert node_count > 0, "No nodes were created!"
    assert edge_count > 0, "No edges were created!"

    print(f"✅ Test passed with {node_count} nodes and {edge_count} edges!")



from typing import Type

from langchain_tests.unit_tests import ToolsUnitTests

from memgraph_toolbox.api.memgraph import Memgraph
from langchain_memgraph.tools import RunQueryTool
from langchain_memgraph.tools import RunShowStorageInfoTool


class TestMemgraphIntegration(ToolsUnitTests):
    @property
    def tool_constructor(self) -> Type[RunQueryTool]:
        return RunQueryTool

    @property
    def tool_constructor_params(self) -> dict:
        # if your tool constructor instead required initialization arguments like
        # `def __init__(self, some_arg: int):`, you would return those here
        # as a dictionary, e.g.: `return {'some_arg': 42}`
        return {"db": Memgraph("bolt://localhost:7687", "", "")}

    @property
    def tool_invoke_params_example(self) -> dict:
        """
        Returns a dictionary representing the "args" of an example tool call.

        This should NOT be a ToolCall dict - i.e. it should not
        have {"name", "id", "args"} keys.
        """
        return {"query": "MATCH (n) RETURN n LIMIT 1"}



"""Memgraph tools."""

from typing import Any, Dict, List, Optional, Type

from langchain_core.callbacks import (
    CallbackManagerForToolRun,
)
from langchain_core.tools import BaseTool
from pydantic import BaseModel, ConfigDict, Field

from memgraph_toolbox.api.memgraph import Memgraph
from memgraph_toolbox.tools.page_rank import PageRankTool
from memgraph_toolbox.tools.config import ShowConfigTool
from memgraph_toolbox.tools.cypher import CypherTool
from memgraph_toolbox.tools.schema import ShowSchemaInfoTool
from memgraph_toolbox.tools.storage import ShowStorageInfoTool
from memgraph_toolbox.tools.trigger import ShowTriggersTool
from memgraph_toolbox.tools.index import ShowIndexInfoTool
from memgraph_toolbox.tools.betweenness_centrality import BetweennessCentralityTool
from memgraph_toolbox.tools.constraint import ShowConstraintInfoTool
from memgraph_toolbox.tools.node_neighborhood import NodeNeighborhoodTool
from memgraph_toolbox.tools.node_vector_search import NodeVectorSearchTool


class BaseMemgraphTool(BaseModel):
    """
    Base tool for interacting with Memgraph.
    """

    db: Memgraph = Field(exclude=True)

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
    )


class _QueryMemgraphToolInput(BaseModel):
    """
    Input query for Memgraph Query tool.
    """

    query: str = Field(..., description="The query to be executed in Memgraph.")


class RunQueryTool(BaseMemgraphTool, BaseTool):
    """Tool for querying Memgraph."""

    name: str = CypherTool(db=None).get_name()
    """The name that is passed to the model when performing tool calling."""

    description: str = CypherTool(db=None).get_description()
    """The description that is passed to the model when performing tool calling."""

    args_schema: Type[BaseModel] = _QueryMemgraphToolInput
    """The schema that is passed to the model when performing tool calling."""

    def _run(
        self,
        query: str,
        run_manager: Optional[CallbackManagerForToolRun] = None,
    ) -> List[Dict[str, Any]]:
        return CypherTool(
            db=self.db,
        ).call({"query": query})


class RunShowSchemaInfoTool(BaseMemgraphTool, BaseTool):
    """Tool for retrieving schema information from Memgraph."""

    name: str = ShowSchemaInfoTool(db=None).get_name()
    """The name that is passed to the model when performing tool calling."""

    description: str = ShowSchemaInfoTool(db=None).get_description()
    """The description that is passed to the model when performing tool calling."""

    args_schema: Optional[Type[BaseModel]] = None
    """The schema that is passed to the model when performing tool calling."""

    def _run(
        self,
        run_manager: Optional[CallbackManagerForToolRun] = None,
    ) -> List[Dict[str, Any]]:
        """Run the tool to get schema information."""

        schema_info = ShowSchemaInfoTool(
            db=self.db,
        )
        result = schema_info.call({})

        return result


class RunShowStorageInfoTool(BaseMemgraphTool, BaseTool):
    """Tool for retrieving storage information from Memgraph."""

    name: str = ShowStorageInfoTool(db=None).get_name()
    """The name that is passed to the model when performing tool calling."""

    description: str = ShowStorageInfoTool(db=None).get_description()
    """The description that is passed to the model when performing tool calling."""

    args_schema: Optional[Type[BaseModel]] = None
    """The schema that is passed to the model when performing tool calling."""

    def _run(
        self,
        run_manager: Optional[CallbackManagerForToolRun] = None,
    ) -> List[Dict[str, Any]]:
        """Run the tool to get storage information."""

        storage_info = ShowStorageInfoTool(
            db=self.db,
        )
        result = storage_info.call({})

        return result


class _PageRankMemgraphToolInput(BaseModel):
    """
    Input schema for the PageRank Memgraph tool.
    """

    limit: int = Field(10, description="Number of nodes to return.")


class RunPageRankMemgraphTool(BaseMemgraphTool, BaseTool):
    """Tool for running PageRank on Memgraph."""

    name: str = PageRankTool(db=None).get_name()
    """The name that is passed to the model when performing tool calling."""

    description: str = PageRankTool(db=None).get_description()
    """The description that is passed to the model when performing tool calling."""

    args_schema: Type[BaseModel] = _PageRankMemgraphToolInput
    """The schema that is passed to the model when performing tool calling."""

    def _run(
        self,
        limit: int = 10,
        run_manager: Optional[CallbackManagerForToolRun] = None,
    ) -> List[Dict[str, Any]]:
        return PageRankTool(
            db=self.db,
        ).call({"limit": limit})


class RunShowConstraintInfoTool(BaseMemgraphTool, BaseTool):
    """Tool for retrieving constraint information from Memgraph."""

    name: str = ShowConstraintInfoTool(db=None).get_name()
    """The name that is passed to the model when performing tool calling."""

    description: str = ShowConstraintInfoTool(db=None).get_description()
    """The description that is passed to the model when performing tool calling."""

    args_schema: Optional[Type[BaseModel]] = None
    """The schema that is passed to the model when performing tool calling."""

    def _run(
        self,
        run_manager: Optional[CallbackManagerForToolRun] = None,
    ) -> List[Dict[str, Any]]:
        """Run the tool to get constraint information."""

        constraint_info = ShowConstraintInfoTool(
            db=self.db,
        )
        result = constraint_info.call({})

        return result


class RunShowIndexInfoTool(BaseMemgraphTool, BaseTool):
    """Tool for retrieving index information from Memgraph."""

    name: str = ShowIndexInfoTool(db=None).get_name()
    """The name that is passed to the model when performing tool calling."""

    description: str = ShowIndexInfoTool(db=None).get_description()
    """The description that is passed to the model when performing tool calling."""

    args_schema: Optional[Type[BaseModel]] = None
    """The schema that is passed to the model when performing tool calling."""

    def _run(
        self,
        run_manager: Optional[CallbackManagerForToolRun] = None,
    ) -> List[Dict[str, Any]]:
        """Run the tool to get index information."""

        index_info = ShowIndexInfoTool(
            db=self.db,
        )
        result = index_info.call({})

        return result


class RunShowConfigTool(BaseMemgraphTool, BaseTool):
    """Tool for retrieving configuration information from Memgraph."""

    name: str = ShowConfigTool(db=None).get_name()
    """The name that is passed to the model when performing tool calling."""

    description: str = ShowConfigTool(db=None).get_description()
    """The description that is passed to the model when performing tool calling."""

    args_schema: Optional[Type[BaseModel]] = None
    """The schema that is passed to the model when performing tool calling."""

    def _run(
        self,
        run_manager: Optional[CallbackManagerForToolRun] = None,
    ) -> List[Dict[str, Any]]:
        """Run the tool to get configuration information."""

        config_info = ShowConfigTool(
            db=self.db,
        )
        result = config_info.call({})

        return result


class RunShowTriggersTool(BaseMemgraphTool, BaseTool):
    """Tool for retrieving trigger information from Memgraph."""

    name: str = ShowTriggersTool(db=None).get_name()
    """The name that is passed to the model when performing tool calling."""

    description: str = ShowTriggersTool(db=None).get_description()
    """The description that is passed to the model when performing tool calling."""

    args_schema: Optional[Type[BaseModel]] = None
    """The schema that is passed to the model when performing tool calling."""

    def _run(
        self,
        run_manager: Optional[CallbackManagerForToolRun] = None,
    ) -> List[Dict[str, Any]]:
        """Run the tool to get trigger information."""

        trigger_info = ShowTriggersTool(
            db=self.db,
        )
        result = trigger_info.call({})

        return result


class _BetweennessCentralityToolInput(BaseModel):
    """
    Input schema for the Betweenness Centrality Memgraph tool.
    """

    isDirectionIgnored: bool = Field(
        True,
        description="Set to false to consider the direction of relationships. Default is true.",
    )
    limit: int = Field(
        10, description="Limit the number of nodes to return. Default is 10."
    )


class RunBetweennessCentralityTool(BaseMemgraphTool, BaseTool):
    """Tool for running Betweenness Centrality on Memgraph."""

    name: str = BetweennessCentralityTool(db=None).get_name()
    """The name that is passed to the model when performing tool calling."""

    description: str = BetweennessCentralityTool(db=None).get_description()
    """The description that is passed to the model when performing tool calling."""

    args_schema: Type[BaseModel] = _BetweennessCentralityToolInput
    """The schema that is passed to the model when performing tool calling."""

    def _run(
        self,
        isDirectionIgnored: bool = True,
        limit: int = 10,
        run_manager: Optional[CallbackManagerForToolRun] = None,
    ) -> List[Dict[str, Any]]:
        return BetweennessCentralityTool(
            db=self.db,
        ).call({"isDirectionIgnored": isDirectionIgnored, "limit": limit})


class _NodeNeighborhoodToolInput(BaseModel):
    """
    Input schema for the Node Neighborhood Memgraph tool.
    """

    node_id: str = Field(
        ...,
        description="The ID of the starting node to find neighborhood around",
    )
    max_distance: int = Field(
        1,
        description="Maximum distance (hops) to search from the starting node. Default is 1.",
    )
    limit: int = Field(
        100, description="Maximum number of nodes to return. Default is 100."
    )


class RunNodeNeighborhoodTool(BaseMemgraphTool, BaseTool):
    """Tool for finding nodes within a specified neighborhood distance in Memgraph."""

    name: str = NodeNeighborhoodTool(db=None).get_name()
    """The name that is passed to the model when performing tool calling."""

    description: str = NodeNeighborhoodTool(db=None).get_description()
    """The description that is passed to the model when performing tool calling."""

    args_schema: Type[BaseModel] = _NodeNeighborhoodToolInput
    """The schema that is passed to the model when performing tool calling."""

    def _run(
        self,
        node_id: str,
        max_distance: int = 1,
        limit: int = 100,
        run_manager: Optional[CallbackManagerForToolRun] = None,
    ) -> List[Dict[str, Any]]:
        return NodeNeighborhoodTool(
            db=self.db,
        ).call({"node_id": node_id, "max_distance": max_distance, "limit": limit})


class _NodeVectorSearchToolInput(BaseModel):
    """
    Input schema for the Node Vector Search Memgraph tool.
    """

    index_name: str = Field(
        ...,
        description="Name of the index to use for the vector search",
    )
    query_vector: List[float] = Field(
        ...,
        description="Query vector to search for similarity",
    )
    limit: int = Field(
        10, description="Number of similar nodes to return. Default is 10."
    )


class RunNodeVectorSearchTool(BaseMemgraphTool, BaseTool):
    """Tool for performing vector similarity search on nodes in Memgraph."""

    name: str = NodeVectorSearchTool(db=None).get_name()
    """The name that is passed to the model when performing tool calling."""

    description: str = NodeVectorSearchTool(db=None).get_description()
    """The description that is passed to the model when performing tool calling."""

    args_schema: Type[BaseModel] = _NodeVectorSearchToolInput
    """The schema that is passed to the model when performing tool calling."""

    def _run(
        self,
        index_name: str,
        query_vector: List[float],
        limit: int = 10,
        run_manager: Optional[CallbackManagerForToolRun] = None,
    ) -> List[Dict[str, Any]]:
        return NodeVectorSearchTool(
            db=self.db,
        ).call({"index_name": index_name, "query_vector": query_vector, "limit": limit})



"""Question answering over a graph."""

from __future__ import annotations

import re
from typing import Any, Dict, List, Optional, Union

from langchain_classic.chains.base import Chain
from langchain_core.callbacks import CallbackManagerForChainRun
from langchain_core.language_models import BaseLanguageModel
from langchain_core.messages import (
    AIMessage,
    BaseMessage,
    SystemMessage,
    ToolMessage,
)
from langchain_core.output_parsers import StrOutputParser
from langchain_core.prompts import (
    BasePromptTemplate,
    ChatPromptTemplate,
    HumanMessagePromptTemplate,
    MessagesPlaceholder,
)
from langchain_core.runnables import Runnable
from pydantic import Field

from langchain_memgraph.chains.prompts import (
    MEMGRAPH_GENERATION_PROMPT,
    MEMGRAPH_QA_PROMPT,
)
from memgraph_toolbox.api.memgraph import Memgraph

INTERMEDIATE_STEPS_KEY = "intermediate_steps"

FUNCTION_RESPONSE_SYSTEM = """You are an assistant that helps to form nice and human 
understandable answers based on the provided information from tools.
Do not add any other information that wasn't present in the tools, and use 
very concise style in interpreting results!
"""


def extract_cypher(text: str) -> str:
    """Extract Cypher code from a text.

    Args:
        text: Text to extract Cypher code from.

    Returns:
        Cypher code extracted from the text.
    """
    # The pattern to find Cypher code enclosed in triple backticks
    pattern = r"```(.*?)```"

    # Find all matches in the input text
    matches = re.findall(pattern, text, re.DOTALL)

    return matches[0] if matches else text


def get_function_response(
    question: str, context: List[Dict[str, Any]]
) -> List[BaseMessage]:
    TOOL_ID = "call_H7fABDuzEau48T10Qn0Lsh0D"
    messages = [
        AIMessage(
            content="",
            additional_kwargs={
                "tool_calls": [
                    {
                        "id": TOOL_ID,
                        "function": {
                            "arguments": '{"question":"' + question + '"}',
                            "name": "GetInformation",
                        },
                        "type": "function",
                    }
                ]
            },
        ),
        ToolMessage(content=str(context), tool_call_id=TOOL_ID),
    ]
    return messages


class MemgraphQAChain(Chain):
    """Chain for question-answering against a graph by generating Cypher statements.

    *Security note*: Make sure that the database connection uses credentials
        that are narrowly-scoped to only include necessary permissions.
        Failure to do so may result in data corruption or loss, since the calling
        code may attempt commands that would result in deletion, mutation
        of data if appropriately prompted or reading sensitive data if such
        data is present in the database.
        The best way to guard against such negative outcomes is to (as appropriate)
        limit the permissions granted to the credentials used with this tool.

        See https://python.langchain.com/docs/security for more information.
    """

    graph: Memgraph = Field(exclude=True)
    cypher_generation_chain: Runnable
    qa_chain: Runnable
    graph_schema: str
    input_key: str = "query"  #: :meta private:
    output_key: str = "result"  #: :meta private:
    top_k: int = 10
    """Number of results to return from the query"""
    return_intermediate_steps: bool = False
    """Whether or not to return the intermediate steps along with the final answer."""
    return_direct: bool = False
    """Optional cypher validation tool"""
    use_function_response: bool = False
    """Whether to wrap the database context as tool/function response"""
    allow_dangerous_requests: bool = False
    """Forced user opt-in to acknowledge that the chain can make dangerous requests.
    
    *Security note*: Make sure that the database connection uses credentials
        that are narrowly-scoped to only include necessary permissions.
        Failure to do so may result in data corruption or loss, since the calling
        code may attempt commands that would result in deletion, mutation
        of data if appropriately prompted or reading sensitive data if such
        data is present in the database.
        The best way to guard against such negative outcomes is to (as appropriate)
        limit the permissions granted to the credentials used with this tool.

        See https://python.langchain.com/docs/security for more information.
    """

    def __init__(self, **kwargs: Any) -> None:
        """Initialize the chain."""
        super().__init__(**kwargs)
        if self.allow_dangerous_requests is not True:
            raise ValueError(
                "In order to use this chain, you must acknowledge that it can make "
                "dangerous requests by setting `allow_dangerous_requests` to `True`."
                "You must narrowly scope the permissions of the database connection "
                "to only include necessary permissions. Failure to do so may result "
                "in data corruption or loss or reading sensitive data if such data is "
                "present in the database."
                "Only use this chain if you understand the risks and have taken the "
                "necessary precautions. "
                "See https://python.langchain.com/docs/security for more information."
            )

    @property
    def input_keys(self) -> List[str]:
        """Return the input keys.

        :meta private:
        """
        return [self.input_key]

    @property
    def output_keys(self) -> List[str]:
        """Return the output keys.

        :meta private:
        """
        _output_keys = [self.output_key]
        return _output_keys

    @property
    def _chain_type(self) -> str:
        return "graph_cypher_chain"

    @classmethod
    def from_llm(
        cls,
        llm: Optional[BaseLanguageModel] = None,
        *,
        qa_prompt: Optional[BasePromptTemplate] = None,
        cypher_prompt: Optional[BasePromptTemplate] = None,
        cypher_llm: Optional[BaseLanguageModel] = None,
        qa_llm: Optional[Union[BaseLanguageModel, Any]] = None,
        qa_llm_kwargs: Optional[Dict[str, Any]] = None,
        cypher_llm_kwargs: Optional[Dict[str, Any]] = None,
        use_function_response: bool = False,
        function_response_system: str = FUNCTION_RESPONSE_SYSTEM,
        **kwargs: Any,
    ) -> MemgraphQAChain:
        """Initialize from LLM."""

        if not cypher_llm and not llm:
            raise ValueError("Either `llm` or `cypher_llm` parameters must be provided")
        if not qa_llm and not llm:
            raise ValueError("Either `llm` or `qa_llm` parameters must be provided")
        if cypher_llm and qa_llm and llm:
            raise ValueError(
                "You can specify up to two of 'cypher_llm', 'qa_llm'"
                ", and 'llm', but not all three simultaneously."
            )
        if cypher_prompt and cypher_llm_kwargs:
            raise ValueError(
                "Specifying cypher_prompt and cypher_llm_kwargs together is"
                " not allowed. Please pass prompt via cypher_llm_kwargs."
            )
        if qa_prompt and qa_llm_kwargs:
            raise ValueError(
                "Specifying qa_prompt and qa_llm_kwargs together is"
                " not allowed. Please pass prompt via qa_llm_kwargs."
            )
        use_qa_llm_kwargs = qa_llm_kwargs if qa_llm_kwargs is not None else {}
        use_cypher_llm_kwargs = (
            cypher_llm_kwargs if cypher_llm_kwargs is not None else {}
        )
        if "prompt" not in use_qa_llm_kwargs:
            use_qa_llm_kwargs["prompt"] = (
                qa_prompt if qa_prompt is not None else MEMGRAPH_QA_PROMPT
            )
        if "prompt" not in use_cypher_llm_kwargs:
            use_cypher_llm_kwargs["prompt"] = (
                cypher_prompt
                if cypher_prompt is not None
                else MEMGRAPH_GENERATION_PROMPT
            )

        qa_llm = qa_llm or llm
        if use_function_response:
            try:
                qa_llm.bind_tools({})  # type: ignore[union-attr]
                response_prompt = ChatPromptTemplate.from_messages(
                    [
                        SystemMessage(content=function_response_system),
                        HumanMessagePromptTemplate.from_template("{question}"),
                        MessagesPlaceholder(variable_name="function_response"),
                    ]
                )
                qa_chain = response_prompt | qa_llm | StrOutputParser()  # type: ignore
            except (NotImplementedError, AttributeError):
                raise ValueError("Provided LLM does not support native tools/functions")
        else:
            qa_chain = use_qa_llm_kwargs["prompt"] | qa_llm | StrOutputParser()  # type: ignore

        prompt = use_cypher_llm_kwargs["prompt"]
        llm_to_use = cypher_llm if cypher_llm is not None else llm

        if prompt is not None and llm_to_use is not None:
            cypher_generation_chain = prompt | llm_to_use | StrOutputParser()  # type: ignore[arg-type]
        else:
            raise ValueError(
                "Missing required components for the cypher generation chain: "
                "'prompt' or 'llm'"
            )

        graph_schema = kwargs["graph"].get_schema

        return cls(
            graph_schema=graph_schema,
            qa_chain=qa_chain,
            cypher_generation_chain=cypher_generation_chain,
            use_function_response=use_function_response,
            **kwargs,
        )

    def _call(
        self,
        inputs: Dict[str, Any],
        run_manager: Optional[CallbackManagerForChainRun] = None,
    ) -> Dict[str, Any]:
        """Generate Cypher statement, use it to look up in db and answer question."""
        _run_manager = run_manager or CallbackManagerForChainRun.get_noop_manager()
        callbacks = _run_manager.get_child()
        question = inputs[self.input_key]
        args = {
            "question": question,
            "schema": self.graph_schema,
        }
        args.update(inputs)

        intermediate_steps: List = []

        generated_cypher = self.cypher_generation_chain.invoke(
            args, callbacks=callbacks
        )
        # Extract Cypher code if it is wrapped in backticks
        generated_cypher = extract_cypher(generated_cypher)

        _run_manager.on_text("Generated Cypher:", end="\n", verbose=self.verbose)
        _run_manager.on_text(
            generated_cypher, color="green", end="\n", verbose=self.verbose
        )

        intermediate_steps.append({"query": generated_cypher})

        # Retrieve and limit the number of results
        # Generated Cypher be null if query corrector identifies invalid schema
        if generated_cypher:
            context = self.graph.query(generated_cypher)[: self.top_k]
        else:
            context = []

        if self.return_direct:
            result = context
        else:
            _run_manager.on_text("Full Context:", end="\n", verbose=self.verbose)
            _run_manager.on_text(
                str(context), color="green", end="\n", verbose=self.verbose
            )

            intermediate_steps.append({"context": context})
            if self.use_function_response:
                function_response = get_function_response(question, context)
                result = self.qa_chain.invoke(  # type: ignore
                    {"question": question, "function_response": function_response},
                )
            else:
                result = self.qa_chain.invoke(  # type: ignore
                    {"question": question, "context": context},
                    callbacks=callbacks,
                )

        chain_result: Dict[str, Any] = {"result": result}
        if self.return_intermediate_steps:
            chain_result[INTERMEDIATE_STEPS_KEY] = intermediate_steps

        return chain_result
